from dune.perftool import Restriction
from dune.perftool.generation import (cached,
                                      domain,
                                      iname,
                                      include_file,
                                      instruction,
                                      symbol,
                                      temporary_variable,
                                      )


@iname
def quadrature_iname():
    domain("q", "q_n")
    return "q"


def quadrature_preamble(code, **kw):
    return instruction(inames=quadrature_iname(), code=code, **kw)


@symbol
def name_quadrature_point():
    return "qp"


def define_quadrature_position(name):
    qp = name_quadrature_point()
    return quadrature_preamble(code="{} = {}.position();".format(name, qp),
                               assignees=frozenset({name}),
                               )


@symbol
def name_quadrature_position():
    name = "pos"
    # To determine the shape, I do query global information here for lack of good alternatives
    from dune.perftool.generation import get_global_context_value
    it = get_global_context_value("integral_type")
    from dune.perftool.pdelab.geometry import name_dimension, name_intersection_dimension
    if it == 'cell':
        shape = (name_dimension(),)
    else:
        shape = (name_intersection_dimension(),)
    temporary_variable(name, shape=shape, shape_impl=("fv",))
    define_quadrature_position(name)
    return name


@symbol
def name_quadrature_position_in_cell(restriction):
    if restriction == Restriction.NONE:
        return name_quadrature_position()
    else:
        from dune.perftool.pdelab.geometry import to_cell_coordinates
        return to_cell_coordinates(name_quadrature_position(), name_quadrature_point(), restriction)


@symbol
def name_quadrature_weight():
    qp = name_quadrature_point()
    return "{}.weight()".format(qp)


def define_quadrature_factor(fac):
    weight = name_quadrature_weight()
    from dune.perftool.pdelab.geometry import name_geometry
    geo = name_geometry()
    pos = name_quadrature_position()
    code = "{} = {}*{}.integrationElement({});".format(fac,
                                                       weight,
                                                       geo,
                                                       pos,
                                                       )
    return quadrature_preamble(code,
                               assignees=frozenset({fac}),
                               read_variables=frozenset({pos}),
                               )


@symbol
def name_factor():
    temporary_variable("fac", shape=())
    define_quadrature_factor("fac")
    return "fac"


@symbol
def name_order():
    # TODO how to determine integration order in a generic fashion here?
    # Or maybe just read it from the ini file? Or define some mapping from the
    # finite elements?
    return "2"


def quadrature_loop_statement():
    include_file('dune/pdelab/common/quadraturerules.hh', filetag='operatorfile')
    qp = name_quadrature_point()
    from dune.perftool.pdelab.geometry import name_geometry
    geo = name_geometry()
    order = name_order()
    return "const auto& {} : quadratureRule({}, {})".format(qp,
                                                            geo,
                                                            order,
                                                            )
